﻿Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput
Imports Autodesk.AutoCAD.GraphicsInterface


Namespace MyBonusPlugin

  Public Class Commands

    ' Class variable to store the instance of our overrule
    Private Shared myOverrule As BillboardAttributesOverrule
    ' Registered Aplication Id for Xdata
    Const mXdataName As String = "ADSK_ATTRIBUTE_ZERO_OVERRULE"

    <CommandMethod("BillboardAttributes")>
    Public Shared Sub ImplementOverrule()

      Dim doc As Document =
            Application.DocumentManager.MdiActiveDocument
      Dim ed As Editor = doc.Editor

      ' Select a block reference

      Dim opts As New PromptEntityOptions(
        vbLf & "Select a block reference:")
      opts.SetRejectMessage(vbLf & "Must be block reference...")
      opts.AddAllowedClass(GetType(BlockReference), True)

      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then
        Exit Sub
      End If

      Dim db As Database = doc.Database

      AddRegAppId(db)

      Using trans As Transaction =
            db.TransactionManager.StartTransaction

        ' Open the BlockReference for read.
        ' We know its a BlockReference because we set a filter in 
        '  our PromptEntityOptions
        Dim blkRef As BlockReference =
                trans.GetObject(res.ObjectId, OpenMode.ForRead)
        ' Record the ObjectIds of all AttributeReferences 
        '  attached to the BlockReference.
        Dim attRefColl As AttributeCollection =
                                          blkRef.AttributeCollection
        ' Iterate through  ObjectIds of all AttributeReferences 
        '  attached to the BlockReference, opening each
        '  AttributeReference and adding xdata to it.
        For Each objId As ObjectId In attRefColl
          Dim attRef As AttributeReference =
                      trans.GetObject(objId, OpenMode.ForWrite)

          ' Create new xdata containing the overrule filter name
          Using resBuf As New ResultBuffer(
            New TypedValue(
             DxfCode.ExtendedDataRegAppName, mXdataName),
            New TypedValue(
             DxfCode.ExtendedDataAsciiString, "Dummy text"))

            ' Add the xdata
            attRef.XData = resBuf
          End Using
        Next
        trans.Commit()

      End Using

      ' Create and register our overrule and turn overruling on.
      ActivateOverrule()

      'Make sure overruling is turned on so our overrule works
      Overrule.Overruling = True

    End Sub


    <CommandMethod("DontBillboardAttributes")>
    Public Shared Sub RemoveXdata()

      Dim doc As Document =
            Application.DocumentManager.MdiActiveDocument
      Dim ed As Editor = doc.Editor

      ' Select a block reference

      Dim opts As New PromptEntityOptions(
        vbLf & "Select a block reference:")
      opts.SetRejectMessage(vbLf & "Must be block reference...")
      opts.AddAllowedClass(GetType(BlockReference), True)

      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then
        Exit Sub
      End If

      Dim db As Database = doc.Database

      AddRegAppId(db)

      Using trans As Transaction =
            db.TransactionManager.StartTransaction

        ' Open the BlockReference for read.
        ' We know its a BlockReference because we set a filter in 
        '  our PromptEntityOptions
        Dim blkRef As BlockReference =
                trans.GetObject(res.ObjectId, OpenMode.ForRead)
        ' Record the ObjectIds of all AttributeReferences 
        '  attached to the BlockReference.
        Dim attRefColl As AttributeCollection =
                                          blkRef.AttributeCollection

        ' Iterate through  ObjectIds of all AttributeReferences 
        '  attached to the BlockReference, opening each
        '  AttributeReference and adding xdata to it.
        For Each objId As ObjectId In attRefColl
          Dim attRef As AttributeReference =
                      trans.GetObject(objId, OpenMode.ForWrite)

          ' Create new xdata containing the overrule filter name
          Using resBuf As New ResultBuffer(
            New TypedValue(
             DxfCode.ExtendedDataRegAppName, mXdataName))
            ' Add the xdata.
            ' Because we leave this blank except for the regappid,
            '  it erases any Xdata we previously added.
            attRef.XData = resBuf
          End Using
        Next
        trans.Commit()

      End Using

    End Sub


    <CommandMethod("ActivateBillboardOverrule")>
    Public Shared Sub ActivateOverrule()

      'We only want to create our overrule instance once, 
      ' so we check if it already exists before we create it
      ' (i.e. this may be the 2nd time we've run the command)
      If myOverrule Is Nothing Then
        'Instantiate our overrule class
        myOverrule = New BillboardAttributesOverrule
        'Register the overrule
        Overrule.AddOverrule(
          RXClass.GetClass(GetType(AttributeReference)),
          myOverrule, False)
      End If

      'Specify which Attributes will be overruled
      myOverrule.SetXDataFilter(mXdataName)
      'Make sure overruling is turned on so our overrule works
      Overrule.Overruling = True
      'Regen drawing to apply overrule effects
      Application.DocumentManager.MdiActiveDocument.Editor.Regen()

    End Sub

    <CommandMethod("DeactivateBillboardOverrule")>
    Public Shared Sub DeActivateOverrule()

      'If we created an overrule we now want to remove it
      If Not myOverrule Is Nothing Then
        'Unregister the overrule
        Overrule.RemoveOverrule(
          RXClass.GetClass(GetType(AttributeReference)),
          myOverrule)
        myOverrule.Dispose()
        myOverrule = Nothing
        'Regen drawing to remove any overrule effects
        Application.DocumentManager.MdiActiveDocument.Editor.Regen()
      End If

    End Sub

    Private Shared Sub AddRegAppId(ByVal db As Database)

      Using trans As Transaction =
        db.TransactionManager.StartTransaction()

        ' First create our RegAppId (if it doesn't already exist)
        Dim appTbl As RegAppTable =
          trans.GetObject(db.RegAppTableId, OpenMode.ForRead)

        If Not appTbl.Has(mXdataName) Then
          Dim appTblRec As New RegAppTableRecord
          appTbl.UpgradeOpen()
          appTblRec.Name = mXdataName
          appTbl.Add(appTblRec)
          trans.AddNewlyCreatedDBObject(appTblRec, True)
        End If
        trans.Commit()
      End Using
    End Sub

  End Class



  'Our custom overrule class derived from DrawableOverrule
  Public Class BillboardAttributesOverrule
    Inherits DrawableOverrule

    'Returning False from WorldDraw tells AutoCAD this entity has
    '  viewport dependent graphics.
    Public Overrides Function WorldDraw(ByVal drawable As Drawable,
                                    ByVal wd As WorldDraw) As Boolean
      Return False
    End Function

    ' Called for each viewport so entity can draw itself differently
    '  depending on the view.
    Public Overrides Sub ViewportDraw(ByVal drawable As Drawable,
                                  ByVal vd As ViewportDraw)
      ' Cast drawable to type AttributeReference (we know it's an
      '  AttributeReference because that's the only class we register our
      '  overrule for).
      Dim attRef As AttributeReference = drawable

      ' Calculate the transformation matrix to rotate from the
      '  attribute's current orientation to the view.
      ' (Thank you to Gile from TheSwamp for correcting this 
      ' transformation to account for insertion on a UCS rotated 
      ' w.r.t. the WCS and for non-left-aligned justification).
      Dim org As Point3d
      If attRef.Justify = AttachmentPoint.BaseLeft Or
        attRef.Justify = AttachmentPoint.BaseAlign Or
        attRef.Justify = AttachmentPoint.BaseFit Then
        org = attRef.Position
      Else
        org = attRef.AlignmentPoint
      End If
      Dim viewMat As Matrix3d =
        Matrix3d.PlaneToWorld(New Plane(org, vd.Viewport.ViewDirection)) *
        Matrix3d.WorldToPlane(New Plane(org, attRef.Normal)) *
        Matrix3d.Rotation(-attRef.Rotation, attRef.Normal, org)

      ' Apply the transformation
      vd.Geometry.PushModelTransform(viewMat)
      'Draw the 'per viewport geometry
      MyBase.ViewportDraw(drawable, vd)
      ' Remove the transformation - we don't want any other objects
      '  to draw themselves in the view plane.
      vd.Geometry.PopModelTransform()

    End Sub


    ' This function tells AutoCAD to dynamically update the 
    '  AttributeReference during view transitions (e.g. 3DORBIT).
    ' Comment it out to improve graphic update performance.
    Public Overrides Function SetAttributes(
                          ByVal drawable As Drawable,
                          ByVal traits As DrawableTraits) As Integer
      Return MyBase.SetAttributes(drawable, traits) Or 2048
    End Function

  End Class


End Namespace